# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: cogment/api/message.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='cogment/api/message.proto',
  package='cogment',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x19\x63ogment/api/message.proto\x12\x07\x63ogment\"J\n\x1cMasterMessageDispatchRequest\x12\x15\n\rsend_to_actor\x18\x01 \x03(\r\x12\x13\n\x0bsend_to_env\x18\x02 \x01(\x08\"\x16\n\x14MessageDispatchReplyb\x06proto3'
)




_MASTERMESSAGEDISPATCHREQUEST = _descriptor.Descriptor(
  name='MasterMessageDispatchRequest',
  full_name='cogment.MasterMessageDispatchRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='send_to_actor', full_name='cogment.MasterMessageDispatchRequest.send_to_actor', index=0,
      number=1, type=13, cpp_type=3, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='send_to_env', full_name='cogment.MasterMessageDispatchRequest.send_to_env', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=38,
  serialized_end=112,
)


_MESSAGEDISPATCHREPLY = _descriptor.Descriptor(
  name='MessageDispatchReply',
  full_name='cogment.MessageDispatchReply',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=114,
  serialized_end=136,
)

DESCRIPTOR.message_types_by_name['MasterMessageDispatchRequest'] = _MASTERMESSAGEDISPATCHREQUEST
DESCRIPTOR.message_types_by_name['MessageDispatchReply'] = _MESSAGEDISPATCHREPLY
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

MasterMessageDispatchRequest = _reflection.GeneratedProtocolMessageType('MasterMessageDispatchRequest', (_message.Message,), {
  'DESCRIPTOR' : _MASTERMESSAGEDISPATCHREQUEST,
  '__module__' : 'cogment.api.message_pb2'
  # @@protoc_insertion_point(class_scope:cogment.MasterMessageDispatchRequest)
  })
_sym_db.RegisterMessage(MasterMessageDispatchRequest)

MessageDispatchReply = _reflection.GeneratedProtocolMessageType('MessageDispatchReply', (_message.Message,), {
  'DESCRIPTOR' : _MESSAGEDISPATCHREPLY,
  '__module__' : 'cogment.api.message_pb2'
  # @@protoc_insertion_point(class_scope:cogment.MessageDispatchReply)
  })
_sym_db.RegisterMessage(MessageDispatchReply)


# @@protoc_insertion_point(module_scope)



# Copyright 2020 Artificial Intelligence Redefined <dev+cogment@ai-r.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import importlib
import logging
from abc import ABC, abstractmethod
from cogment.session import Session


_OBSERVATION = "observation"
_FINAL_DATA = "final_data"
_MESSAGE = "message"
_REWARD = "reward"


class ActorClass:
    def __init__(
        self,
        id,
        config_type,
        action_space,
        observation_space,
        observation_delta,
        observation_delta_apply_fn,
        feedback_space,
    ):
        self.id = id
        self.config_type = config_type
        self.action_space = action_space
        self.observation_space = observation_space
        self.observation_delta = observation_delta
        self.observation_delta_apply_fn = observation_delta_apply_fn
        self.feedback_space = feedback_space


class ActorClassList:
    def __init__(self, *args):
        self._actor_classes_list = list(args)

        for a_c in args:
            setattr(self, a_c.id, a_c)

    def __iter__(self):
        return iter(self._actor_classes_list)

    def __contains__(self, key):
        return hasattr(self, key)

    def __getitem__(self, key):
        return getattr(self, key)

    def get_class_by_index(self, index):
        return self._actor_classes_list[index]


class Reward:
    def __init__(self):
        self.tick_id = -1
        self.value = 0
        self.confidence = 0

        self._feedbacks = None

    def _set_all(self, reward, tick_id):
        self.tick_id = tick_id
        self.value = reward.value
        self.confidence = reward.confidence
        self._set_feedbacks(reward.feedbacks)

    def _set_feedbacks(self, feedbacks):
        self._feedbacks = feedbacks
        if self.tick_id == -1 and self._feedbacks:
            self.tick_id = self._feedbacks[0].tick_id

    def all_user_data(self):
        assert self._feedbacks
        for fdbk in self._feedbacks:
            assert fdbk.tick_id == self.tick_id
            if fdbk.content:
                yield fdbk.content


class ActorSession(Session):
    """This represents an actor being performed locally."""

    def __init__(self, impl, actor_class, trial, name, impl_name, config):
        super().__init__(trial)
        self.actor_class = actor_class
        self.name = name
        self.impl_name = impl_name
        self.config = config

        self._ended = False
        self._action_queue = asyncio.Queue()
        self._task = None  # Task used to call _run()
        self._latest_observation = None
        self._last_event_received = False

        self.__impl = impl
        self.__started = False
        self.__event_queue = None
        self.__obs_future = None

    @abstractmethod
    async def _retrieve_action(self):
        pass

    def start(self):
        assert not self.__started
        assert not self._ended

        self.__event_queue = asyncio.Queue()
        self.__started = True

    async def event_loop(self):
        assert self.__started
        assert not self._ended

        loop_active = not self._last_event_received
        while loop_active:
            event = await self.__event_queue.get()
            self._last_event_received = _FINAL_DATA in event
            keep_looping = yield event
            self.__event_queue.task_done()
            loop_active = (keep_looping is None or bool(keep_looping)) and not self._last_event_received

    def do_action(self, action):
        assert self.__started
        self._action_queue.put_nowait(action)

    async def _end(self, package):
        if not self._ended:
            self._ended = True

        if self.__event_queue:
            std_messages = []
            for msg in package.messages:
                std_messages.append((msg.sender_name(), msg.payload()))
            package.messages = std_messages

            event = {_FINAL_DATA : package}
            await self.__event_queue.put(event)
        else:
            logging.warning("The actor received final data that it was unable to handle.")

    def _new_observation(self, obs):
        if not self.__started or self._ended:
            return

        if self.__event_queue:
            event = {_OBSERVATION : obs}
            self.__event_queue.put_nowait(event)
        else:
            logging.warning("The actor received an observation that it was unable to handle.")

    def _new_reward(self, reward):
        if not self.__started or self._ended:
            return

        if self.__event_queue:
            event = {_REWARD : reward}
            self.__event_queue.put_nowait(event)
        else:
            logging.warning("The actor received a reward that it was unable to handle.")

    def _new_message(self, message):
        if not self.__started or self._ended:
            return

        if self.__event_queue:
            event = {_MESSAGE : (message.sender_name(), message.payload())}
            self.__event_queue.put_nowait(event)
        else:
            logging.warning("The actor received a message that it was unable to handle.")

    async def _run(self):
        await self.__impl(self)


class _ServedActorSession(ActorSession):
    def __init__(self, impl, actor_class, trial, name, impl_name, config):
        super().__init__(impl, actor_class, trial, name, impl_name, config)

    async def _retrieve_action(self):
        action = await self._action_queue.get()
        self._action_queue.task_done()
        return action


class _ClientActorSession(ActorSession):
    def __init__(self, impl, actor_class, trial, name, impl_name, config):
        super().__init__(impl, actor_class, trial, name, impl_name, config)

    async def _retrieve_action(self):
        action = await self._action_queue.get()
        self._action_queue.task_done()
        return action



# Copyright 2020 Artificial Intelligence Redefined <dev+cogment@ai-r.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from cogment.api.common_pb2 import VersionInfo, ObservationData, TrialParams
from cogment.version import __version__
from cogment.delta_encoding import DecodeObservationData

from types import SimpleNamespace

import importlib
import grpc


def list_versions():
    reply = VersionInfo()
    reply.versions.add(name='cogment_sdk', version=__version__)
    reply.versions.add(name='grpc', version=grpc.__version__)

    return reply


def raw_params_to_user_params(params, settings):
    trial_config = None
    if params.HasField("trial_config"):
        trial_config = settings.trial.config_type()
        trial_config.ParseFromString(params.trial_config.content)

    env_config = None
    if(params.environment.HasField("config")):
        env_config = settings.environment.config_type()
        env_config.ParseFromString(params.environment.config.content)

    environment = SimpleNamespace(
        endpoint=params.environment.endpoint,
        config=env_config
    )

    actors = []
    for actor in params.actors:
        actor_config = None

        if actor.HasField("config"):
            a_c = settings.actor_classes.__getattribute__(actor.actor_class)
            actor_config = a_c.config_type()
            actor_config.ParseFromString(actor.config.content)

        actor_data = SimpleNamespace(
            actor_class=actor.actor_class,
            endpoint=actor.endpoint,
            config=actor_config
        )
        actors.append(actor_data)

    return SimpleNamespace(
        trial_config=trial_config,
        environment=environment,
        actors=actors,
        max_steps=params.max_steps,
        max_inactivity=params.max_inactivity
    )


def user_params_to_raw_params(params, settings):
    result = TrialParams()

    result.max_steps = params.max_steps
    result.max_inactivity = params.max_inactivity

    if params.trial_config is not None:
        result.trial_config.content = params.trial_config.SerializeToString()

    result.environment.endpoint = params.environment.endpoint
    if params.environment.config is not None:
        result.environment.config.content = \
            params.environment.config.SerializeToString()

    for actor_data in params.actors:
        actor_pb = result.actors.add()
        actor_pb.actor_class = actor_data.actor_class
        actor_pb.endpoint = actor_data.endpoint
        if actor_data.config is not None:
            actor_pb.config.content = actor_data.config.SerializeToString()

    return result


class DecodeData():

    def __init__(self, trial_params, cog_settings):
        self.__cog_settings = cog_settings
        self.last_obs = []

        actor_classes_list = [
            actor.id for actor in self.__cog_settings.actor_classes]
        trial_actor_list = [actor.actor_class for actor in trial_params.actors]
        self.actor_counts = [0] * len(actor_classes_list)
        for index, actor_class in enumerate(actor_classes_list):
            self.actor_counts[index] += trial_actor_list.count(actor_class)

        for ac_index, actor_class in enumerate(self.__cog_settings.actor_classes):
            count = self.actor_counts[ac_index]
            self.last_obs.extend([None] * count)

    def decode_datasample(self, sample):

        actor_index = 0
        for ac_index, actor_class in enumerate(self.__cog_settings.actor_classes):
            count = self.actor_counts[ac_index]
            for _ in range(count):
                try:
                    obs_id = sample.observations.actors_map[actor_index]
                except Exception:
                    print(sample)

                obs_data = sample.observations.observations[obs_id]

                obs = DecodeObservationData(
                    actor_class, obs_data, self.last_obs[actor_index])
                self.last_obs[actor_index] = obs

                actor_index += 1

        action_list = []
        for act_data in sample.actions:
            action = self.__cog_settings.data_pb.Action()
            action.ParseFromString(act_data.content)
            action_list.append(action)

        reward_list = []
        for rwd in sample.rewards:
            reward_list.append((rwd.value, rwd.confidence))

        message_list = []
        for messages in sample.messages:
            sub_msg_list = []
            for message in messages.messages:

                class_type = message.payload.type_url.split('.')
                user_data = getattr(importlib.import_module(
                    self.__cog_settings.protolib), class_type[-1])()
                message.payload.Unpack(user_data)

                sub_msg_list.append((message.sender_name, user_data))

            message_list.append(sub_msg_list)

        return self.last_obs, action_list, reward_list, message_list



# Copyright 2020 Artificial Intelligence Redefined <dev+cogment@ai-r.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod


class PrehookSession(ABC):
    """This represents the session for a prehook for a trial"""

    def __init__(self, params, trial):
        self._params = params
        self._trial = trial

        self.trial_config = params.trial_config
        self.trial_max_steps = params.max_steps
        self.trial_max_inactivity = params.max_inactivity

        self.environment_config = params.environment.config
        self.environment_endpoint = params.environment.endpoint

        self.actors_config = [actor.config for actor in params.actors]
        self.actors_endpoint = [actor.endpoint for actor in params.actors]
        self.actors_class = [actor.actor_class for actor in params.actors]

    @abstractmethod
    def _recode(self):
        pass

    def get_trial_id(self):
        return self._trial.id

    def validate(self):
        self._recode()


class _ServedPrehookSession(PrehookSession):
    def __init__(self, params, trial):
        super().__init__(params, trial)

    def _recode(self):
        self._params.trial_config = self.trial_config
        self._params.max_steps = self.trial_max_steps
        self._params.max_inactivity = self.trial_max_inactivity

        self._params.environment.config = self.environment_config
        self._params.environment.endpoint = self.environment_endpoint

        nb_actors = len(self._params.actors)
        if len(self.actors_config) != nb_actors:
            raise AttributeError("Wrong number of actor configurations")
        if len(self.actors_endpoint) != nb_actors:
            raise AttributeError("Wrong number of actor endpoints")
        if len(self.actors_class) != nb_actors:
            raise AttributeError("Wrong number of actor classes")

        for index in range(nb_actors):
            self._params.actors[index].config = self.actors_config[index]
            self._params.actors[index].endpoint = self.actors_endpoint[index]
            self._params.actors[index].actor_class = self.actors_class[index]



# Copyright 2020 Artificial Intelligence Redefined <dev+cogment@ai-r.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import grpc
import grpc.experimental.aio

import cogment.api.orchestrator_pb2 as orchestrator
from cogment.api.orchestrator_pb2_grpc import TrialLifecycleStub
from cogment.trial import Trial
from cogment.control import _ServedControlSession


class ControlServicer:
    def __init__(self, cog_settings, endpoint):
        self.cog_settings = cog_settings
        channel = grpc.experimental.aio.insecure_channel(endpoint)
        self.lifecycle_stub = TrialLifecycleStub(channel)

    async def run(self, user_id, impl, trial_config):
        req = orchestrator.TrialStartRequest()
        req.user_id = user_id
        if trial_config is not None:
            req.config.content = trial_config.SerializeToString()

        # TODO: after the next line the process hangs indefinitely when running two integration tests
        rep = await self.lifecycle_stub.StartTrial(req)
        trial = Trial(rep.trial_id, rep.actors_in_trial, self.cog_settings)

        await impl(_ServedControlSession(trial, self.lifecycle_stub))



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from cogment.api import common_pb2 as cogment_dot_api_dot_common__pb2
from cogment.api import hooks_pb2 as cogment_dot_api_dot_hooks__pb2


class TrialHooksStub(object):
    """Expected metadata for all calls except `Version`:
    - trial-id: The id of the trial
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.OnPreTrial = channel.unary_unary(
                '/cogment.TrialHooks/OnPreTrial',
                request_serializer=cogment_dot_api_dot_hooks__pb2.PreTrialContext.SerializeToString,
                response_deserializer=cogment_dot_api_dot_hooks__pb2.PreTrialContext.FromString,
                )
        self.Version = channel.unary_unary(
                '/cogment.TrialHooks/Version',
                request_serializer=cogment_dot_api_dot_common__pb2.VersionRequest.SerializeToString,
                response_deserializer=cogment_dot_api_dot_common__pb2.VersionInfo.FromString,
                )


class TrialHooksServicer(object):
    """Expected metadata for all calls except `Version`:
    - trial-id: The id of the trial
    """

    def OnPreTrial(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Version(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_TrialHooksServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'OnPreTrial': grpc.unary_unary_rpc_method_handler(
                    servicer.OnPreTrial,
                    request_deserializer=cogment_dot_api_dot_hooks__pb2.PreTrialContext.FromString,
                    response_serializer=cogment_dot_api_dot_hooks__pb2.PreTrialContext.SerializeToString,
            ),
            'Version': grpc.unary_unary_rpc_method_handler(
                    servicer.Version,
                    request_deserializer=cogment_dot_api_dot_common__pb2.VersionRequest.FromString,
                    response_serializer=cogment_dot_api_dot_common__pb2.VersionInfo.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'cogment.TrialHooks', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class TrialHooks(object):
    """Expected metadata for all calls except `Version`:
    - trial-id: The id of the trial
    """

    @staticmethod
    def OnPreTrial(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/cogment.TrialHooks/OnPreTrial',
            cogment_dot_api_dot_hooks__pb2.PreTrialContext.SerializeToString,
            cogment_dot_api_dot_hooks__pb2.PreTrialContext.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Version(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/cogment.TrialHooks/Version',
            cogment_dot_api_dot_common__pb2.VersionRequest.SerializeToString,
            cogment_dot_api_dot_common__pb2.VersionInfo.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



# Copyright 2020 Artificial Intelligence Redefined <dev+cogment@ai-r.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from threading import Thread
from typing import Callable, Awaitable, Dict, List, Any
from types import ModuleType
import os
import asyncio
from traceback import print_exc
from types import SimpleNamespace
import grpc
import grpc.experimental.aio
from prometheus_client import start_http_server, CollectorRegistry

from cogment.actor import ActorSession, ActorClass
from cogment.environment import EnvironmentSession
from cogment.prehook import PrehookSession
from cogment.datalog import DatalogSession
from cogment.control import ControlSession

# Agent
from cogment.agent_service import AgentServicer
from cogment.api.agent_pb2 import _AGENTENDPOINT as agent_endpoint_descriptor
from cogment.api.agent_pb2_grpc import add_AgentEndpointServicer_to_server

# Client
from cogment.client_service import ClientServicer

# Control
from cogment.control_service import ControlServicer

# Environment
from cogment.env_service import EnvironmentServicer
from cogment.api.environment_pb2_grpc import add_EnvironmentEndpointServicer_to_server

# Prehook
from cogment.hooks_service import PrehookServicer
from cogment.api.hooks_pb2_grpc import add_TrialHooksServicer_to_server

# Log Exporter
from cogment.log_exporter_service import LogExporterService
from cogment.api.data_pb2_grpc import add_LogExporterServicer_to_server


DEFAULT_MAX_WORKERS = 1
ENABLE_REFLECTION_VAR_NAME = "COGMENT_GRPC_REFLECTION"
DEFAULT_ENABLE_REFLECTION = os.getenv(ENABLE_REFLECTION_VAR_NAME, "false")
DEFAULT_PROMETHEUS_PORT = 8000


def _add_actor_service(grpc_server, impls, service_names, cog_settings, prometheus_registry):
    servicer = AgentServicer(impls, cog_settings, prometheus_registry)
    add_AgentEndpointServicer_to_server(servicer, grpc_server)
    service_names.append(agent_endpoint_descriptor.full_name)


def _add_env_service(grpc_server, impls, cog_settings, prometheus_registry):
    servicer = EnvironmentServicer(impls, cog_settings, prometheus_registry)
    add_EnvironmentEndpointServicer_to_server(servicer, grpc_server)


def _add_prehook_service(grpc_server, impls, cog_settings, prometheus_registry):
    servicer = PrehookServicer(impls, cog_settings, prometheus_registry)
    add_TrialHooksServicer_to_server(servicer, grpc_server)


def _add_datalog_service(grpc_server, impl, cog_settings, prometheus_registry):
    servicer = LogExporterService(impl, cog_settings, prometheus_registry)
    add_LogExporterServicer_to_server(servicer, grpc_server)


class Context:
    def __init__(self, user_id: str, cog_settings: ModuleType):
        self._user_id = user_id
        self.__actor_impls: Dict[str, SimpleNamespace] = {}
        self.__env_impls: Dict[str, SimpleNamespace] = {}
        self.__prehook_impls: List[Callable[[PrehookSession], Awaitable[None]]] = []
        self.__datalog_impl: Callable[[DatalogSession], Awaitable[None]] = None
        self._grpc_server = None  # type: Any
        self._prometheus_registry = CollectorRegistry()
        self.__cog_settings = cog_settings

    def register_actor(self,
                       impl: Callable[[ActorSession], Awaitable[None]],
                       impl_name: str,
                       actor_classes: List[str] = []):
        assert impl_name not in self.__actor_impls
        assert self._grpc_server is None

        self.__actor_impls[impl_name] = SimpleNamespace(
            impl=impl, actor_classes=actor_classes
        )

    def register_environment(self,
                             impl: Callable[[EnvironmentSession], Awaitable[None]],
                             impl_name: str = "default"):
        assert impl_name not in self.__env_impls
        assert self._grpc_server is None

        self.__env_impls[impl_name] = SimpleNamespace(impl=impl)

    def register_pre_trial_hook(self,
                                impl: Callable[[PrehookSession], Awaitable[None]]):
        assert self._grpc_server is None

        self.__prehook_impls.append(impl)

    def register_datalog(self,
                         impl: Callable[[DatalogSession], Awaitable[None]]):
        assert self._grpc_server is None
        assert self.__datalog_impl is None

        self.__datalog_impl = impl

    async def serve_all_registered(self,
                                   port: int,
                                   prometheus_port: int = DEFAULT_PROMETHEUS_PORT):
        try:
            start_http_server(prometheus_port)

            if self.__actor_impls or self.__env_impls or self.__prehook_impls or self.__datalog_impl is not None:
                self._grpc_server = grpc.experimental.aio.server()

                service_names: List[str] = []
                if self.__actor_impls:
                    _add_actor_service(
                        self._grpc_server,
                        self.__actor_impls,
                        service_names,
                        self.__cog_settings,
                        self._prometheus_registry
                    )

                if self.__env_impls:
                    _add_env_service(
                        self._grpc_server,
                        self.__env_impls,
                        self.__cog_settings,
                        self._prometheus_registry
                    )

                if self.__prehook_impls:
                    _add_prehook_service(
                        self._grpc_server,
                        self.__prehook_impls,
                        self.__cog_settings,
                        self._prometheus_registry
                    )

                if self.__datalog_impl is not None:
                    _add_datalog_service(
                        self._grpc_server,
                        self.__datalog_impl,
                        self.__cog_settings,
                        self._prometheus_registry
                    )

                self._grpc_server.add_insecure_port(f"[::]:{port}")
                await self._grpc_server.start()
                await self._grpc_server.wait_for_termination()
        except Exception as e:
            print_exc()
            raise e

    async def start_trial(self,
                          endpoint,
                          impl: Callable[[ControlSession], Awaitable[None]],
                          trial_config=None):
        try:
            servicer = ControlServicer(self.__cog_settings, endpoint)
            await servicer.run(self._user_id, impl, trial_config)
        except Exception as e:
            print_exc()
            raise e

    async def join_trial(self, trial_id, endpoint, impl_name, actor_name=None):
        actor_impl = self.__actor_impls[impl_name]

        servicer = ClientServicer(self.__cog_settings, endpoint)
        await servicer.run(trial_id, actor_impl.impl, impl_name, actor_impl.actor_classes, actor_name)



# Copyright 2020 Artificial Intelligence Redefined <dev+cogment@ai-r.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import grpc
import grpc.experimental.aio

import cogment.api.orchestrator_pb2 as orchestrator
from cogment.actor import _ClientActorSession, Reward
from cogment.api.common_pb2 import TrialActor
from cogment.api.orchestrator_pb2_grpc import ActorEndpointStub
from cogment.delta_encoding import DecodeObservationData
from cogment.errors import InvalidRequestError
from cogment.trial import Trial
from types import SimpleNamespace

import asyncio


async def read_observations(client_session, action_conn):
    while True:
        reply = await action_conn.read()

        if not reply.final_data:
            for obs_reply in reply.data.observations:
                tick_id = obs_reply.tick_id
                client_session._trial.tick_id = tick_id

                obs = DecodeObservationData(
                    client_session.actor_class,
                    obs_reply.data,
                    client_session._latest_observation)
                client_session._latest_observation = obs
                client_session._new_observation(obs)

            for rew_request in reply.data.rewards():
                reward = Reward()
                reward._set_all(rew_request, -1)
                client_session._new_reward(reward)

            for message in reply.data.messages:
                client_session._new_message(message)

        else:
            client_session._trial.over = True

            package = SimpleNamespace(observations=[], rewards=[], messages=[])
            tick_id = None

            for obs_reply in reply.data.observations():
                tick_id = obs_reply.tick_id
                obs = DecodeObservationData(
                    client_session.actor_class,
                    obs_reply.data,
                    client_session._latest_observation)
                client_session._latest_observation = obs
                package.observations.append(obs)

            for rew_request in reply.data.rewards():
                reward = Reward()
                reward._set_all(rew_request, -1)
                package.rewards.append(reward)

            for msg_request in reply.data.messages():
                package.messages.append(msg_request)

            if tick_id is not None:
                client_session._trial.tick_id = tick_id

            await client_session._end(package)

            break


async def write_actions(client_session, action_conn):
    while True:
        act = await client_session._retrieve_action()
        action_req = orchestrator.TrialActionRequest()

        if act is not None:
            action_req.action.content = act.SerializeToString()

        await action_conn.write(action_req)


class ClientServicer:
    def __init__(self, cog_settings, endpoint):
        self.cog_settings = cog_settings

        channel = grpc.experimental.aio.insecure_channel(endpoint)
        self._actor_stub = ActorEndpointStub(channel)

    async def run(self, trial_id, impl, impl_name, actor_classes, actor_name):

        # TODO: Handle properly the multiple actor classes.  Including "all" classes
        #       when the list is empty
        if actor_name is None:
            requested_actor_class = actor_classes[0]
            req = orchestrator.TrialJoinRequest(trial_id=trial_id, actor_class=requested_actor_class)
        else:
            req = orchestrator.TrialJoinRequest(trial_id=trial_id, actor_name=actor_name)

        reply = await self._actor_stub.JoinTrial(req)

        trial = Trial(reply.trial_id, reply.actors_in_trial, self.cog_settings)

        self_info = None
        for info in reply.actors_in_trial:
            if info.name == reply.actor_name:
                self_info = info
                break
        if self_info is None:
            raise InvalidRequestError(f"Unknown actor name: {reply.actor_name}", request=reply)

        actor_class = self.cog_settings.actor_classes[self_info.actor_class]

        config = None
        if reply.HasField("config"):
            if actor_class.config_type is None:
                raise Exception(
                    f"Actor [{self_info.name}] received config data of unknown type (was it defined in cogment.yaml)")
            config = actor_class.config_type()
            config.ParseFromString(reply.config.content)

        new_session = _ClientActorSession(
            impl, actor_class, trial, self_info.name, impl_name, config
        )

        loop = asyncio.get_running_loop()

        metadata = (("trial-id", trial.id), ("actor-name", self_info.name))
        action_conn = self._actor_stub.ActionStream(metadata)

        reader_task = loop.create_task(read_observations(new_session, action_conn))
        writer_task = loop.create_task(write_actions(new_session, action_conn))

        await impl(new_session)

        reader_task.cancel()
        writer_task.cancel()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: cogment/api/orchestrator.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from cogment.api import common_pb2 as cogment_dot_api_dot_common__pb2
from cogment.api import agent_pb2 as cogment_dot_api_dot_agent__pb2
from cogment.api import environment_pb2 as cogment_dot_api_dot_environment__pb2
from cogment.api import message_pb2 as cogment_dot_api_dot_message__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='cogment/api/orchestrator.proto',
  package='cogment',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x1e\x63ogment/api/orchestrator.proto\x12\x07\x63ogment\x1a\x18\x63ogment/api/common.proto\x1a\x17\x63ogment/api/agent.proto\x1a\x1d\x63ogment/api/environment.proto\x1a\x19\x63ogment/api/message.proto\"J\n\x11TrialStartRequest\x12$\n\x06\x63onfig\x18\x01 \x01(\x0b\x32\x14.cogment.TrialConfig\x12\x0f\n\x07user_id\x18\x02 \x01(\t\"Q\n\x0fTrialStartReply\x12\x10\n\x08trial_id\x18\x01 \x01(\t\x12,\n\x0f\x61\x63tors_in_trial\x18\x02 \x03(\x0b\x32\x13.cogment.TrialActor\"\x17\n\x15TerminateTrialRequest\"\x15\n\x13TerminateTrialReply\"2\n\x10TrialInfoRequest\x12\x1e\n\x16get_latest_observation\x18\x01 \x01(\x08\"v\n\tTrialInfo\x12\x10\n\x08trial_id\x18\x01 \x01(\t\x12\"\n\x05state\x18\x02 \x01(\x0e\x32\x13.cogment.TrialState\x12\x33\n\x12latest_observation\x18\x03 \x01(\x0b\x32\x17.cogment.ObservationSet\"3\n\x0eTrialInfoReply\x12!\n\x05trial\x18\x01 \x03(\x0b\x32\x12.cogment.TrialInfo\"c\n\x10TrialJoinRequest\x12\x10\n\x08trial_id\x18\x01 \x01(\t\x12\x15\n\x0b\x61\x63tor_class\x18\x02 \x01(\tH\x00\x12\x14\n\nactor_name\x18\x03 \x01(\tH\x00\x42\x10\n\x0eslot_selection\"\x8a\x01\n\x0eTrialJoinReply\x12\x12\n\nactor_name\x18\x01 \x01(\t\x12\x10\n\x08trial_id\x18\x02 \x01(\t\x12$\n\x06\x63onfig\x18\x03 \x01(\x0b\x32\x14.cogment.ActorConfig\x12,\n\x0f\x61\x63tors_in_trial\x18\x04 \x03(\x0b\x32\x13.cogment.TrialActor\"\x17\n\x15TrialHeartbeatRequest\"\x15\n\x13TrialHeartbeatReply\"5\n\x12TrialActionRequest\x12\x1f\n\x06\x61\x63tion\x18\x01 \x01(\x0b\x32\x0f.cogment.Action\"N\n\x10TrialActionReply\x12&\n\x04\x64\x61ta\x18\x01 \x01(\x0b\x32\x18.cogment.AgentPeriodData\x12\x12\n\nfinal_data\x18\x02 \x01(\x08\"<\n\x14TrialFeedbackRequest\x12$\n\tfeedbacks\x18\x01 \x03(\x0b\x32\x11.cogment.Feedback\"\x14\n\x12TrialFeedbackReply\"9\n\x13TrialMessageRequest\x12\"\n\x08messages\x18\x01 \x03(\x0b\x32\x10.cogment.Message\"\x13\n\x11TrialMessageReply*a\n\nTrialState\x12\x0b\n\x07UNKNOWN\x10\x00\x12\x10\n\x0cINITIALIZING\x10\x01\x12\x0b\n\x07PENDING\x10\x02\x12\x0b\n\x07RUNNING\x10\x03\x12\x0f\n\x0bTERMINATING\x10\x04\x12\t\n\x05\x45NDED\x10\x05\x32\x81\x03\n\x0eTrialLifecycle\x12\x44\n\nStartTrial\x12\x1a.cogment.TrialStartRequest\x1a\x18.cogment.TrialStartReply\"\x00\x12P\n\x0eTerminateTrial\x12\x1e.cogment.TerminateTrialRequest\x1a\x1c.cogment.TerminateTrialReply\"\x00\x12U\n\x0bSendMessage\x12%.cogment.MasterMessageDispatchRequest\x1a\x1d.cogment.MessageDispatchReply\"\x00\x12\x44\n\x0cGetTrialInfo\x12\x19.cogment.TrialInfoRequest\x1a\x17.cogment.TrialInfoReply\"\x00\x12:\n\x07Version\x12\x17.cogment.VersionRequest\x1a\x14.cogment.VersionInfo\"\x00\x32\xc2\x03\n\rActorEndpoint\x12\x41\n\tJoinTrial\x12\x19.cogment.TrialJoinRequest\x1a\x17.cogment.TrialJoinReply\"\x00\x12L\n\x0c\x41\x63tionStream\x12\x1b.cogment.TrialActionRequest\x1a\x19.cogment.TrialActionReply\"\x00(\x01\x30\x01\x12K\n\tHeartbeat\x12\x1e.cogment.TrialHeartbeatRequest\x1a\x1c.cogment.TrialHeartbeatReply\"\x00\x12L\n\x0cGiveFeedback\x12\x1d.cogment.TrialFeedbackRequest\x1a\x1b.cogment.TrialFeedbackReply\"\x00\x12I\n\x0bSendMessage\x12\x1c.cogment.TrialMessageRequest\x1a\x1a.cogment.TrialMessageReply\"\x00\x12:\n\x07Version\x12\x17.cogment.VersionRequest\x1a\x14.cogment.VersionInfo\"\x00\x62\x06proto3'
  ,
  dependencies=[cogment_dot_api_dot_common__pb2.DESCRIPTOR,cogment_dot_api_dot_agent__pb2.DESCRIPTOR,cogment_dot_api_dot_environment__pb2.DESCRIPTOR,cogment_dot_api_dot_message__pb2.DESCRIPTOR,])

_TRIALSTATE = _descriptor.EnumDescriptor(
  name='TrialState',
  full_name='cogment.TrialState',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='UNKNOWN', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='INITIALIZING', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PENDING', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='RUNNING', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='TERMINATING', index=4, number=4,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='ENDED', index=5, number=5,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=1173,
  serialized_end=1270,
)
_sym_db.RegisterEnumDescriptor(_TRIALSTATE)

TrialState = enum_type_wrapper.EnumTypeWrapper(_TRIALSTATE)
UNKNOWN = 0
INITIALIZING = 1
PENDING = 2
RUNNING = 3
TERMINATING = 4
ENDED = 5



_TRIALSTARTREQUEST = _descriptor.Descriptor(
  name='TrialStartRequest',
  full_name='cogment.TrialStartRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='config', full_name='cogment.TrialStartRequest.config', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='user_id', full_name='cogment.TrialStartRequest.user_id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=152,
  serialized_end=226,
)


_TRIALSTARTREPLY = _descriptor.Descriptor(
  name='TrialStartReply',
  full_name='cogment.TrialStartReply',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='trial_id', full_name='cogment.TrialStartReply.trial_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='actors_in_trial', full_name='cogment.TrialStartReply.actors_in_trial', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=228,
  serialized_end=309,
)


_TERMINATETRIALREQUEST = _descriptor.Descriptor(
  name='TerminateTrialRequest',
  full_name='cogment.TerminateTrialRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=311,
  serialized_end=334,
)


_TERMINATETRIALREPLY = _descriptor.Descriptor(
  name='TerminateTrialReply',
  full_name='cogment.TerminateTrialReply',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=336,
  serialized_end=357,
)


_TRIALINFOREQUEST = _descriptor.Descriptor(
  name='TrialInfoRequest',
  full_name='cogment.TrialInfoRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='get_latest_observation', full_name='cogment.TrialInfoRequest.get_latest_observation', index=0,
      number=1, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=359,
  serialized_end=409,
)


_TRIALINFO = _descriptor.Descriptor(
  name='TrialInfo',
  full_name='cogment.TrialInfo',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='trial_id', full_name='cogment.TrialInfo.trial_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='state', full_name='cogment.TrialInfo.state', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='latest_observation', full_name='cogment.TrialInfo.latest_observation', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=411,
  serialized_end=529,
)


_TRIALINFOREPLY = _descriptor.Descriptor(
  name='TrialInfoReply',
  full_name='cogment.TrialInfoReply',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='trial', full_name='cogment.TrialInfoReply.trial', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=531,
  serialized_end=582,
)


_TRIALJOINREQUEST = _descriptor.Descriptor(
  name='TrialJoinRequest',
  full_name='cogment.TrialJoinRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='trial_id', full_name='cogment.TrialJoinRequest.trial_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='actor_class', full_name='cogment.TrialJoinRequest.actor_class', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='actor_name', full_name='cogment.TrialJoinRequest.actor_name', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='slot_selection', full_name='cogment.TrialJoinRequest.slot_selection',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=584,
  serialized_end=683,
)


_TRIALJOINREPLY = _descriptor.Descriptor(
  name='TrialJoinReply',
  full_name='cogment.TrialJoinReply',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='actor_name', full_name='cogment.TrialJoinReply.actor_name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='trial_id', full_name='cogment.TrialJoinReply.trial_id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='config', full_name='cogment.TrialJoinReply.config', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='actors_in_trial', full_name='cogment.TrialJoinReply.actors_in_trial', index=3,
      number=4, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=686,
  serialized_end=824,
)


_TRIALHEARTBEATREQUEST = _descriptor.Descriptor(
  name='TrialHeartbeatRequest',
  full_name='cogment.TrialHeartbeatRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=826,
  serialized_end=849,
)


_TRIALHEARTBEATREPLY = _descriptor.Descriptor(
  name='TrialHeartbeatReply',
  full_name='cogment.TrialHeartbeatReply',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=851,
  serialized_end=872,
)


_TRIALACTIONREQUEST = _descriptor.Descriptor(
  name='TrialActionRequest',
  full_name='cogment.TrialActionRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='action', full_name='cogment.TrialActionRequest.action', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=874,
  serialized_end=927,
)


_TRIALACTIONREPLY = _descriptor.Descriptor(
  name='TrialActionReply',
  full_name='cogment.TrialActionReply',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='data', full_name='cogment.TrialActionReply.data', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='final_data', full_name='cogment.TrialActionReply.final_data', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=929,
  serialized_end=1007,
)


_TRIALFEEDBACKREQUEST = _descriptor.Descriptor(
  name='TrialFeedbackRequest',
  full_name='cogment.TrialFeedbackRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='feedbacks', full_name='cogment.TrialFeedbackRequest.feedbacks', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1009,
  serialized_end=1069,
)


_TRIALFEEDBACKREPLY = _descriptor.Descriptor(
  name='TrialFeedbackReply',
  full_name='cogment.TrialFeedbackReply',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1071,
  serialized_end=1091,
)


_TRIALMESSAGEREQUEST = _descriptor.Descriptor(
  name='TrialMessageRequest',
  full_name='cogment.TrialMessageRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='messages', full_name='cogment.TrialMessageRequest.messages', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1093,
  serialized_end=1150,
)


_TRIALMESSAGEREPLY = _descriptor.Descriptor(
  name='TrialMessageReply',
  full_name='cogment.TrialMessageReply',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1152,
  serialized_end=1171,
)

_TRIALSTARTREQUEST.fields_by_name['config'].message_type = cogment_dot_api_dot_common__pb2._TRIALCONFIG
_TRIALSTARTREPLY.fields_by_name['actors_in_trial'].message_type = cogment_dot_api_dot_common__pb2._TRIALACTOR
_TRIALINFO.fields_by_name['state'].enum_type = _TRIALSTATE
_TRIALINFO.fields_by_name['latest_observation'].message_type = cogment_dot_api_dot_environment__pb2._OBSERVATIONSET
_TRIALINFOREPLY.fields_by_name['trial'].message_type = _TRIALINFO
_TRIALJOINREQUEST.oneofs_by_name['slot_selection'].fields.append(
  _TRIALJOINREQUEST.fields_by_name['actor_class'])
_TRIALJOINREQUEST.fields_by_name['actor_class'].containing_oneof = _TRIALJOINREQUEST.oneofs_by_name['slot_selection']
_TRIALJOINREQUEST.oneofs_by_name['slot_selection'].fields.append(
  _TRIALJOINREQUEST.fields_by_name['actor_name'])
_TRIALJOINREQUEST.fields_by_name['actor_name'].containing_oneof = _TRIALJOINREQUEST.oneofs_by_name['slot_selection']
_TRIALJOINREPLY.fields_by_name['config'].message_type = cogment_dot_api_dot_common__pb2._ACTORCONFIG
_TRIALJOINREPLY.fields_by_name['actors_in_trial'].message_type = cogment_dot_api_dot_common__pb2._TRIALACTOR
_TRIALACTIONREQUEST.fields_by_name['action'].message_type = cogment_dot_api_dot_common__pb2._ACTION
_TRIALACTIONREPLY.fields_by_name['data'].message_type = cogment_dot_api_dot_agent__pb2._AGENTPERIODDATA
_TRIALFEEDBACKREQUEST.fields_by_name['feedbacks'].message_type = cogment_dot_api_dot_common__pb2._FEEDBACK
_TRIALMESSAGEREQUEST.fields_by_name['messages'].message_type = cogment_dot_api_dot_common__pb2._MESSAGE
DESCRIPTOR.message_types_by_name['TrialStartRequest'] = _TRIALSTARTREQUEST
DESCRIPTOR.message_types_by_name['TrialStartReply'] = _TRIALSTARTREPLY
DESCRIPTOR.message_types_by_name['TerminateTrialRequest'] = _TERMINATETRIALREQUEST
DESCRIPTOR.message_types_by_name['TerminateTrialReply'] = _TERMINATETRIALREPLY
DESCRIPTOR.message_types_by_name['TrialInfoRequest'] = _TRIALINFOREQUEST
DESCRIPTOR.message_types_by_name['TrialInfo'] = _TRIALINFO
DESCRIPTOR.message_types_by_name['TrialInfoReply'] = _TRIALINFOREPLY
DESCRIPTOR.message_types_by_name['TrialJoinRequest'] = _TRIALJOINREQUEST
DESCRIPTOR.message_types_by_name['TrialJoinReply'] = _TRIALJOINREPLY
DESCRIPTOR.message_types_by_name['TrialHeartbeatRequest'] = _TRIALHEARTBEATREQUEST
DESCRIPTOR.message_types_by_name['TrialHeartbeatReply'] = _TRIALHEARTBEATREPLY
DESCRIPTOR.message_types_by_name['TrialActionRequest'] = _TRIALACTIONREQUEST
DESCRIPTOR.message_types_by_name['TrialActionReply'] = _TRIALACTIONREPLY
DESCRIPTOR.message_types_by_name['TrialFeedbackRequest'] = _TRIALFEEDBACKREQUEST
DESCRIPTOR.message_types_by_name['TrialFeedbackReply'] = _TRIALFEEDBACKREPLY
DESCRIPTOR.message_types_by_name['TrialMessageRequest'] = _TRIALMESSAGEREQUEST
DESCRIPTOR.message_types_by_name['TrialMessageReply'] = _TRIALMESSAGEREPLY
DESCRIPTOR.enum_types_by_name['TrialState'] = _TRIALSTATE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

TrialStartRequest = _reflection.GeneratedProtocolMessageType('TrialStartRequest', (_message.Message,), {
  'DESCRIPTOR' : _TRIALSTARTREQUEST,
  '__module__' : 'cogment.api.orchestrator_pb2'
  # @@protoc_insertion_point(class_scope:cogment.TrialStartRequest)
  })
_sym_db.RegisterMessage(TrialStartRequest)

TrialStartReply = _reflection.GeneratedProtocolMessageType('TrialStartReply', (_message.Message,), {
  'DESCRIPTOR' : _TRIALSTARTREPLY,
  '__module__' : 'cogment.api.orchestrator_pb2'
  # @@protoc_insertion_point(class_scope:cogment.TrialStartReply)
  })
_sym_db.RegisterMessage(TrialStartReply)

TerminateTrialRequest = _reflection.GeneratedProtocolMessageType('TerminateTrialRequest', (_message.Message,), {
  'DESCRIPTOR' : _TERMINATETRIALREQUEST,
  '__module__' : 'cogment.api.orchestrator_pb2'
  # @@protoc_insertion_point(class_scope:cogment.TerminateTrialRequest)
  })
_sym_db.RegisterMessage(TerminateTrialRequest)

TerminateTrialReply = _reflection.GeneratedProtocolMessageType('TerminateTrialReply', (_message.Message,), {
  'DESCRIPTOR' : _TERMINATETRIALREPLY,
  '__module__' : 'cogment.api.orchestrator_pb2'
  # @@protoc_insertion_point(class_scope:cogment.TerminateTrialReply)
  })
_sym_db.RegisterMessage(TerminateTrialReply)

TrialInfoRequest = _reflection.GeneratedProtocolMessageType('TrialInfoRequest', (_message.Message,), {
  'DESCRIPTOR' : _TRIALINFOREQUEST,
  '__module__' : 'cogment.api.orchestrator_pb2'
  # @@protoc_insertion_point(class_scope:cogment.TrialInfoRequest)
  })
_sym_db.RegisterMessage(TrialInfoRequest)

TrialInfo = _reflection.GeneratedProtocolMessageType('TrialInfo', (_message.Message,), {
  'DESCRIPTOR' : _TRIALINFO,
  '__module__' : 'cogment.api.orchestrator_pb2'
  # @@protoc_insertion_point(class_scope:cogment.TrialInfo)
  })
_sym_db.RegisterMessage(TrialInfo)

TrialInfoReply = _reflection.GeneratedProtocolMessageType('TrialInfoReply', (_message.Message,), {
  'DESCRIPTOR' : _TRIALINFOREPLY,
  '__module__' : 'cogment.api.orchestrator_pb2'
  # @@protoc_insertion_point(class_scope:cogment.TrialInfoReply)
  })
_sym_db.RegisterMessage(TrialInfoReply)

TrialJoinRequest = _reflection.GeneratedProtocolMessageType('TrialJoinRequest', (_message.Message,), {
  'DESCRIPTOR' : _TRIALJOINREQUEST,
  '__module__' : 'cogment.api.orchestrator_pb2'
  # @@protoc_insertion_point(class_scope:cogment.TrialJoinRequest)
  })
_sym_db.RegisterMessage(TrialJoinRequest)

TrialJoinReply = _reflection.GeneratedProtocolMessageType('TrialJoinReply', (_message.Message,), {
  'DESCRIPTOR' : _TRIALJOINREPLY,
  '__module__' : 'cogment.api.orchestrator_pb2'
  # @@protoc_insertion_point(class_scope:cogment.TrialJoinReply)
  })
_sym_db.RegisterMessage(TrialJoinReply)

TrialHeartbeatRequest = _reflection.GeneratedProtocolMessageType('TrialHeartbeatRequest', (_message.Message,), {
  'DESCRIPTOR' : _TRIALHEARTBEATREQUEST,
  '__module__' : 'cogment.api.orchestrator_pb2'
  # @@protoc_insertion_point(class_scope:cogment.TrialHeartbeatRequest)
  })
_sym_db.RegisterMessage(TrialHeartbeatRequest)

TrialHeartbeatReply = _reflection.GeneratedProtocolMessageType('TrialHeartbeatReply', (_message.Message,), {
  'DESCRIPTOR' : _TRIALHEARTBEATREPLY,
  '__module__' : 'cogment.api.orchestrator_pb2'
  # @@protoc_insertion_point(class_scope:cogment.TrialHeartbeatReply)
  })
_sym_db.RegisterMessage(TrialHeartbeatReply)

TrialActionRequest = _reflection.GeneratedProtocolMessageType('TrialActionRequest', (_message.Message,), {
  'DESCRIPTOR' : _TRIALACTIONREQUEST,
  '__module__' : 'cogment.api.orchestrator_pb2'
  # @@protoc_insertion_point(class_scope:cogment.TrialActionRequest)
  })
_sym_db.RegisterMessage(TrialActionRequest)

TrialActionReply = _reflection.GeneratedProtocolMessageType('TrialActionReply', (_message.Message,), {
  'DESCRIPTOR' : _TRIALACTIONREPLY,
  '__module__' : 'cogment.api.orchestrator_pb2'
  # @@protoc_insertion_point(class_scope:cogment.TrialActionReply)
  })
_sym_db.RegisterMessage(TrialActionReply)

TrialFeedbackRequest = _reflection.GeneratedProtocolMessageType('TrialFeedbackRequest', (_message.Message,), {
  'DESCRIPTOR' : _TRIALFEEDBACKREQUEST,
  '__module__' : 'cogment.api.orchestrator_pb2'
  # @@protoc_insertion_point(class_scope:cogment.TrialFeedbackRequest)
  })
_sym_db.RegisterMessage(TrialFeedbackRequest)

TrialFeedbackReply = _reflection.GeneratedProtocolMessageType('TrialFeedbackReply', (_message.Message,), {
  'DESCRIPTOR' : _TRIALFEEDBACKREPLY,
  '__module__' : 'cogment.api.orchestrator_pb2'
  # @@protoc_insertion_point(class_scope:cogment.TrialFeedbackReply)
  })
_sym_db.RegisterMessage(TrialFeedbackReply)

TrialMessageRequest = _reflection.GeneratedProtocolMessageType('TrialMessageRequest', (_message.Message,), {
  'DESCRIPTOR' : _TRIALMESSAGEREQUEST,
  '__module__' : 'cogment.api.orchestrator_pb2'
  # @@protoc_insertion_point(class_scope:cogment.TrialMessageRequest)
  })
_sym_db.RegisterMessage(TrialMessageRequest)

TrialMessageReply = _reflection.GeneratedProtocolMessageType('TrialMessageReply', (_message.Message,), {
  'DESCRIPTOR' : _TRIALMESSAGEREPLY,
  '__module__' : 'cogment.api.orchestrator_pb2'
  # @@protoc_insertion_point(class_scope:cogment.TrialMessageReply)
  })
_sym_db.RegisterMessage(TrialMessageReply)



_TRIALLIFECYCLE = _descriptor.ServiceDescriptor(
  name='TrialLifecycle',
  full_name='cogment.TrialLifecycle',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_start=1273,
  serialized_end=1658,
  methods=[
  _descriptor.MethodDescriptor(
    name='StartTrial',
    full_name='cogment.TrialLifecycle.StartTrial',
    index=0,
    containing_service=None,
    input_type=_TRIALSTARTREQUEST,
    output_type=_TRIALSTARTREPLY,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='TerminateTrial',
    full_name='cogment.TrialLifecycle.TerminateTrial',
    index=1,
    containing_service=None,
    input_type=_TERMINATETRIALREQUEST,
    output_type=_TERMINATETRIALREPLY,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='SendMessage',
    full_name='cogment.TrialLifecycle.SendMessage',
    index=2,
    containing_service=None,
    input_type=cogment_dot_api_dot_message__pb2._MASTERMESSAGEDISPATCHREQUEST,
    output_type=cogment_dot_api_dot_message__pb2._MESSAGEDISPATCHREPLY,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetTrialInfo',
    full_name='cogment.TrialLifecycle.GetTrialInfo',
    index=3,
    containing_service=None,
    input_type=_TRIALINFOREQUEST,
    output_type=_TRIALINFOREPLY,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='Version',
    full_name='cogment.TrialLifecycle.Version',
    index=4,
    containing_service=None,
    input_type=cogment_dot_api_dot_common__pb2._VERSIONREQUEST,
    output_type=cogment_dot_api_dot_common__pb2._VERSIONINFO,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
])
_sym_db.RegisterServiceDescriptor(_TRIALLIFECYCLE)

DESCRIPTOR.services_by_name['TrialLifecycle'] = _TRIALLIFECYCLE


_ACTORENDPOINT = _descriptor.ServiceDescriptor(
  name='ActorEndpoint',
  full_name='cogment.ActorEndpoint',
  file=DESCRIPTOR,
  index=1,
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_start=1661,
  serialized_end=2111,
  methods=[
  _descriptor.MethodDescriptor(
    name='JoinTrial',
    full_name='cogment.ActorEndpoint.JoinTrial',
    index=0,
    containing_service=None,
    input_type=_TRIALJOINREQUEST,
    output_type=_TRIALJOINREPLY,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ActionStream',
    full_name='cogment.ActorEndpoint.ActionStream',
    index=1,
    containing_service=None,
    input_type=_TRIALACTIONREQUEST,
    output_type=_TRIALACTIONREPLY,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='Heartbeat',
    full_name='cogment.ActorEndpoint.Heartbeat',
    index=2,
    containing_service=None,
    input_type=_TRIALHEARTBEATREQUEST,
    output_type=_TRIALHEARTBEATREPLY,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GiveFeedback',
    full_name='cogment.ActorEndpoint.GiveFeedback',
    index=3,
    containing_service=None,
    input_type=_TRIALFEEDBACKREQUEST,
    output_type=_TRIALFEEDBACKREPLY,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='SendMessage',
    full_name='cogment.ActorEndpoint.SendMessage',
    index=4,
    containing_service=None,
    input_type=_TRIALMESSAGEREQUEST,
    output_type=_TRIALMESSAGEREPLY,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='Version',
    full_name='cogment.ActorEndpoint.Version',
    index=5,
    containing_service=None,
    input_type=cogment_dot_api_dot_common__pb2._VERSIONREQUEST,
    output_type=cogment_dot_api_dot_common__pb2._VERSIONINFO,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
])
_sym_db.RegisterServiceDescriptor(_ACTORENDPOINT)

DESCRIPTOR.services_by_name['ActorEndpoint'] = _ACTORENDPOINT

# @@protoc_insertion_point(module_scope)



# Copyright 2020 Artificial Intelligence Redefined <dev+cogment@ai-r.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod

from cogment.errors import InvalidRequestError
from cogment.api.environment_pb2_grpc import EnvironmentEndpointServicer

import cogment.api.environment_pb2 as env_api
from cogment.api.common_pb2 import Feedback, ObservationData, Observation
from cogment.utils import list_versions

from types import SimpleNamespace, ModuleType
from typing import Any, Dict, Tuple
import grpc.experimental.aio

from cogment.environment import _ServedEnvironmentSession, ENVIRONMENT_ACTOR_NAME

from cogment.trial import Trial

from prometheus_client import Summary, Counter

import traceback
import atexit
import logging
import typing
import asyncio
from time import time
from traceback import print_exc


def new_actions_table(settings, trial):
    actions_by_actor_class = settings.ActionsTable(trial)
    actions_by_actor_id = actions_by_actor_class.all_actions()

    return actions_by_actor_class, actions_by_actor_id


def pack_observations(env_session, observations, reply, tick_id):
    timestamp = int(time() * 1000000000)

    new_obs = [None] * len(env_session._trial.actors)

    for tgt, obs in observations:
        for actor_index, actor in enumerate(env_session._trial.actors):
            if isinstance(tgt, int):
                new_obs[actor_index] = obs
            elif tgt == "*" or tgt == "*.*":
                new_obs[actor_index] = obs
            elif isinstance(tgt, str):
                if "." not in tgt:
                    if actor.name == tgt:
                        new_obs[actor_index] = obs
                else:
                    class_name = tgt.split(".")
                    if class_name[1] == actor.name:
                        new_obs[actor_index] = obs
                    elif class_name[1] == "*":
                        if actor.actor_class.id == class_name[0]:
                            new_obs[actor_index] = obs

    snapshots = [True] * len(env_session._trial.actors)

    for actor_index, actor in enumerate(env_session._trial.actors):
        if not new_obs[actor_index]:
            raise Exception("An actor is missing an observation")
        snapshots[actor_index] = isinstance(
            new_obs[actor_index], actor.actor_class.observation_space
        )

    # dupping time
    seen_observations = {}

    reply.observation_set.tick_id = tick_id
    reply.observation_set.timestamp = timestamp

    for actor_index, actor in enumerate(env_session._trial.actors):
        obs_id = id(new_obs[actor_index])
        obs_key = seen_observations.get(obs_id)
        if obs_key is None:
            obs_key = len(reply.observation_set.observations)

            observation_data = ObservationData(
                content=new_obs[actor_index].SerializeToString(),
                snapshot=snapshots[actor_index],
            )

            reply.observation_set.observations.append(observation_data)

            seen_observations[obs_id] = obs_key

        reply.observation_set.actors_map.append(obs_key)


def _process_reply(observations, env_session):
    rep = env_api.EnvActionReply()

    rep.feedbacks.extend(env_session._trial._gather_all_feedback())
    rep.messages.extend(env_session._trial._gather_all_messages(ENVIRONMENT_ACTOR_NAME))
    pack_observations(env_session, observations, rep, env_session._trial.tick_id)

    return rep


async def write_observations(context, env_session):
    try:
        while True:
            observations, final = await env_session._retrieve_obs()

            reply = _process_reply(observations, env_session)
            reply.final_update = final
            await context.write(reply)

            if final:
                break
    except Exception as e:
        print_exc()
        raise e


def _process_actions(request, env_session):
    len_actions = len(request.action_set.actions)
    len_actors = len(env_session._trial._actions_by_actor_id)

    if len_actions != len_actors:
        raise Exception(
            f"Received {len_actions} actions but have {len_actors} actors"
        )

    for i, action in enumerate(env_session._trial._actions_by_actor_id):
        action.ParseFromString(request.action_set.actions[i])

    return env_session._trial._actions


async def read_actions(context, env_session):
    while True:
        request = await context.read()

        if request == grpc.experimental.aio.EOF:
            break

        env_session._new_action(_process_actions(request, env_session))


class EnvironmentServicer(EnvironmentEndpointServicer):
    def __init__(self, env_impls, cog_settings, prometheus_registry):
        self.__impls = env_impls
        self.__env_sessions = {}
        self.__cog_settings = cog_settings

        self.UPDATE_REQUEST_TIME = Summary(
            "environment_update_processing_seconds",
            "Times spend by an environment on the update function",
            ["impl_name"],
            registry=prometheus_registry
        )
        self.TRAINING_DURATION = Summary(
            "environment_trial_duration", "Trial duration", ["trial_actor"],
            registry=prometheus_registry
        )
        self.TRIALS_STARTED = Counter(
            "environment_trials_started", "Number of trial starts", ["impl_name"],
            registry=prometheus_registry
        )
        self.TRIALS_ENDED = Counter(
            "environment_trials_ended", "Number of trial ends", ["impl_name"],
            registry=prometheus_registry
        )
        self.MESSAGES_RECEIVED = Counter(
            "environment_received_messages",
            "Number of messages received",
            ["impl_name"],
            registry=prometheus_registry
        )

        atexit.register(self.__cleanup)

        logging.info("Environment Service started")

    async def OnStart(self, request, context):
        metadata = dict(context.invocation_metadata())
        trial_id = metadata["trial-id"]

        target_impl_name = request.impl_name
        if not target_impl_name:
            target_impl_name = "default"

        impl = self.__impls.get(target_impl_name)
        if impl is None:
            raise InvalidRequestError(
                message=f"Unknown environment implementation name [{target_impl_name}]", request=request
            )

        key = trial_id
        if key in self.__env_sessions:
            raise InvalidRequestError(
                message=f"Environment already exists for trial [{trial_id}]", request=request
            )

        self.TRIALS_STARTED.labels(target_impl_name).inc()

        trial = Trial(trial_id, request.actors_in_trial, self.__cog_settings)
        trial.tick_id = 0

        # action table time
        actions_by_actor_class, actions_by_actor_id = new_actions_table(self.__cog_settings, trial)

        trial._actions = actions_by_actor_class
        trial._actions_by_actor_id = actions_by_actor_id

        config = None
        if request.HasField("config"):
            if self.__cog_settings.environment.config_type is None:
                raise Exception("Environment received config data of unknown type (was it defined in cogment.yaml)")
            config = self.__cog_settings.environment.config_type()
            config.ParseFromString(request.config.content)

        new_session = _ServedEnvironmentSession(impl.impl, trial, target_impl_name, config)
        self.__env_sessions[key] = new_session

        loop = asyncio.get_running_loop()
        new_session._task = loop.create_task(new_session._run())

        observations, _ = await new_session._retrieve_obs()

        reply = env_api.EnvStartReply()
        pack_observations(new_session, observations, reply, 0)

        return reply

    async def OnAction(self, request_iterator, context):
        try:
            metadata = dict(context.invocation_metadata())
            trial_id = metadata["trial-id"]

            key = trial_id
            env_session = self.__env_sessions[key]

            with self.UPDATE_REQUEST_TIME.labels(env_session.impl_name).time():
                env_session._trial.tick_id += 1

                # We are going to have three concurrent coroutines:
                # - One that reads actions from the orchestrator
                # - One that sends observations to the orchestrator
                # - The environment's main (which will be the current coroutine)
                loop = asyncio.get_running_loop()
                reader_task = loop.create_task(read_actions(context, env_session))
                writer_task = loop.create_task(write_observations(context, env_session))

                await env_session._task
                if not env_session._ended:
                    del self.__env_sessions[key]
                    raise Exception(f"Trial [{trial_id}] was never ended")

                await writer_task
                await reader_task

                del self.__env_sessions[key]
        except Exception as e:
            print_exc()
            raise e

    async def OnMessage(self, request, context):
        metadata = dict(context.invocation_metadata())
        trial_id = metadata["trial-id"]

        key = trial_id
        env_session = self.__env_sessions[key]

        for message in request.messages:
            self.MESSAGES_RECEIVED.labels(env_session.impl_name).inc()
            env_session._new_message(message)

        return env_api.EnvMessageReply()

    async def OnEnd(self, request, context):
        metadata = dict(context.invocation_metadata())
        trial_id = metadata["trial-id"]

        key = trial_id
        env_session = self.__env_sessions[key]

        for actor in env_session._trial.actors:
            self.TRAINING_DURATION.labels(actor.name).observe(env_session._trial.tick_id)

        obs = await env_session._end_request(_process_actions(request, env_session))

        if obs:
            reply = _process_reply(obs, env_session)
        else:
            reply = env_api.EnvActionReply()
        reply.set_end_trial(True)

        self.TRIALS_ENDED.labels(env_session.impl_name).inc()
        self.__env_sessions.pop(key, None)

        return reply

    def __cleanup(self):
        self.__env_sessions.clear()
        atexit.unregister(self.__cleanup)



# Copyright 2020 Artificial Intelligence Redefined <dev+cogment@ai-r.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from cogment.api.agent_pb2_grpc import AgentEndpointServicer
import cogment.api.agent_pb2 as agent_api

from cogment.utils import list_versions
from cogment.trial import Trial

from cogment.errors import InvalidRequestError
from cogment.delta_encoding import DecodeObservationData
from cogment.actor import _ServedActorSession, Reward

from prometheus_client import Summary, Counter, Gauge

from types import SimpleNamespace
import traceback
import atexit
import logging
import typing
import asyncio
from traceback import print_exc
import grpc.experimental.aio


def _trial_key(trial_id, actor_name):
    return f"{trial_id}_{actor_name}"


def _impl_can_serve_actor_class(impl, actor_class):
    if impl.actor_classes:
        for ac in impl.actor_classes:
            if ac == actor_class.id:
                return True
        return False
    else:
        return True


async def read_observations(context, agent_session):
    try:
        while True:
            request = await context.read()

            # This means the GRPC channel has been closed by the orchestrator.
            if request == grpc.experimental.aio.EOF:
                break

            agent_session._trial.tick_id = request.observation.tick_id

            obs = DecodeObservationData(
                agent_session.actor_class,
                request.observation.data,
                agent_session._latest_observation,
            )
            agent_session._latest_observation = obs
            agent_session._new_observation(obs)
    except Exception as e:
        print_exc()
        raise e


async def write_actions(context, agent_session):
    try:
        while True:
            act = await agent_session._retrieve_action()
            msg = agent_api.AgentActionReply()

            if act is not None:
                msg.action.content = act.SerializeToString()

            msg.feedbacks.extend(agent_session._trial._gather_all_feedback())

            actor_name = dict(context.invocation_metadata())["actor-name"]
            msg.messages.extend(agent_session._trial._gather_all_messages(actor_name))

            await context.write(msg)
    except Exception as e:
        print_exc()
        raise e


class AgentServicer(AgentEndpointServicer):
    def __init__(self, agent_impls, cog_settings, prometheus_registry):
        self.__impls = agent_impls
        self.__agent_sessions = {}
        self.__cog_settings = cog_settings
        atexit.register(self.__cleanup)

        self.DECIDE_REQUEST_TIME = Summary(
            "actor_decide_processing_seconds",
            "Time spent by an actor on the decide function",
            ["name"],
            registry=prometheus_registry
        )
        self.ACTORS_STARTED = Counter(
            "actor_started", "Number of actors created", ["impl_name"],
            registry=prometheus_registry
        )
        self.ACTORS_ENDED = Counter(
            "actor_ended", "Number of actors ended", ["impl_name"],
            registry=prometheus_registry
        )
        self.MESSAGES_RECEIVED = Counter(
            "actor_received_messages", "Number of messages received", ["name"],
            registry=prometheus_registry
        )
        self.REWARDS_RECEIVED = Gauge(
            "actor_reward_summation", "Cumulative rewards received", ["name"],
            registry=prometheus_registry
        )

        logging.info("Agent Service started")

    async def OnStart(self, request, context):
        metadata = dict(context.invocation_metadata())
        actor_name = str(metadata["actor-name"])
        trial_id = metadata["trial-id"]

        key = _trial_key(trial_id, actor_name)

        if request.impl_name not in self.__impls:
            raise InvalidRequestError(
                message=f"Unknown agent impl: {request.impl_name}", request=request
            )
        impl = self.__impls[request.impl_name]

        self_info = None
        for info in request.actors_in_trial:
            if info.name == actor_name:
                self_info = info
                break
        if self_info is None:
            raise InvalidRequestError(f"Unknown agent name: {actor_name}", request=request)

        if self_info.actor_class not in self.__cog_settings.actor_classes:
            raise InvalidRequestError(
                message=f"Unknown agent class: {request.actor_class}", request=request
            )
        actor_class = self.__cog_settings.actor_classes[self_info.actor_class]

        if not _impl_can_serve_actor_class(impl, actor_class):
            raise InvalidRequestError(
                message=f"{request.impl_name} does not implement {request.actor_class}",
                request=request,
            )

        if key in self.__agent_sessions:
            raise InvalidRequestError(message="Agent already exists", request=request)

        self.ACTORS_STARTED.labels(request.impl_name).inc()

        trial = Trial(trial_id, request.actors_in_trial, self.__cog_settings)

        config = None
        if request.HasField("config"):
            if actor_class.config_type is None:
                raise Exception(
                    f"Actor [{actor_name}] received config data of unknown type (was it defined in cogment.yaml)")
            config = actor_class.config_type()
            config.ParseFromString(request.config.content)

        new_session = _ServedActorSession(
            impl.impl, actor_class, trial, self_info.name, request.impl_name, config
        )
        self.__agent_sessions[key] = new_session

        loop = asyncio.get_running_loop()
        new_session._task = loop.create_task(new_session._run())

        return agent_api.AgentStartReply()

    async def OnEnd(self, request, context):
        metadata = dict(context.invocation_metadata())
        key = _trial_key(metadata["trial-id"], metadata["actor-name"])
        agent_session = self.__agent_sessions[key]

        package = SimpleNamespace(observations=[], rewards=[], messages=[])
        for obs_request in request.final_data.observations:
            obs = DecodeObservationData(
                agent_session.actor_class,
                obs_request.data,
                agent_session._latest_observation)
            agent_session._latest_observation = obs
            package.observations.append(obs)

        for rew_request in request.final_data.rewards:
            reward = Reward()
            reward._set_all(rew_request, -1)
            package.rewards.append(reward)

        for msg_request in request.final_data.messages:
            package.messages.append(msg_request)

        await agent_session._end(package)

        self.ACTORS_ENDED.labels(agent_session.impl_name).inc()

        self.__agent_sessions.pop(key, None)

        return agent_api.AgentEndReply()

    async def OnObservation(self, request_iterator, context):
        metadata = dict(context.invocation_metadata())
        key = _trial_key(metadata["trial-id"], metadata["actor-name"])

        agent_session = self.__agent_sessions[key]

        with self.DECIDE_REQUEST_TIME.labels(agent_session.name).time():
            loop = asyncio.get_running_loop()

            reader_task = loop.create_task(read_observations(context, agent_session))
            writer_task = loop.create_task(write_actions(context, agent_session))

            await agent_session._task

            reader_task.cancel()
            writer_task.cancel()

            del self.__agent_sessions[key]

    async def OnReward(self, request, context):
        metadata = dict(context.invocation_metadata())
        actor_name = str(metadata["actor-name"])
        trial_id = metadata["trial-id"]

        key = _trial_key(trial_id, actor_name)

        reward = Reward()
        reward._set_all(request.reward, request.tick_id)

        agent_session = self.__agent_sessions[key]
        if reward.value < 0.0:
            self.REWARDS_RECEIVED.labels(agent_session.name).dec(abs(reward.value))
        else:
            self.REWARDS_RECEIVED.labels(agent_session.name).inc(reward.value)

        agent_session._new_reward(reward)

        return agent_api.AgentRewardReply()

    async def OnMessage(self, request, context):
        metadata = dict(context.invocation_metadata())
        actor_name = str(metadata["actor-name"])
        trial_id = metadata["trial-id"]

        key = _trial_key(trial_id, actor_name)

        agent_session = self.__agent_sessions[key]

        for message in request.messages:
            agent_session._new_message(message)
            self.MESSAGES_RECEIVED.labels(agent_session.name).inc()

        return agent_api.AgentMessageReply()

    async def Version(self, request, context):
        try:
            return list_versions()
        except Exception:
            traceback.print_exc()
            raise

    def __cleanup(self):
        self.__agent_sessions.clear()
        atexit.unregister(self.__cleanup)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['cogment']

package_data = \
{'': ['*']}

install_requires = \
['grpcio-reflection',
 'grpcio==1.30',
 'prometheus-client>=0.8.0,<0.9.0',
 'protobuf>=3.7']

setup_kwargs = {
    'name': 'cogment',
    'version': '1.0.0a2',
    'description': 'Cogment python SDK',
    'long_description': '# cogment python SDK\n\n## Introduction\n\nThe Cogment framework is a high-efficiency, open source framework designed to enable the training of models in environments where humans and agents interact with the environment and each other continuously. It’s capable of distributed, multi-agent, multi-model training.\n\nThis is the python API for making use of the cogment framework when working with the Python programming language.\n\nFor further Cogment information, check out the documentation at <https://docs.cogment.ai>\n\n## Developers\n\n### Local setup\n\nMake sure you have the following installed:\n  - [Python](https://www.python.org) (any version >3.8 should work),\n  - [Poetry](https://python-poetry.org).\n\nInstall the dependencies, including downloading and building the cogment protobuf API, by navigating to the python SDK directory and run the following\n\n```\npoetry install\n```\n\n### Define used Cogment protobuf API\n\nThe version of the used cogment protobuf API is defined in the `.cogment-api.yaml` file at the root of the repository. The following can be defined:\n\n- `cogment_api_version: "latest"`, is the default, it retrieves the _latest_ build of the cogment-api `develop`,\n- `cogment_api_version: "vMAJOR.MINOR.PATCH[-PRERELEASE]"`, retrieves an official release of cogment-api.\n- `cogment_api_path: "../path/to/cogment-api"`, retrieves a local version of cogment-api found at the given path ; if set, this overrides `cogment_api_version`.\n\n> ⚠️ when building a docker image, `cogment_api_path` needs to exists in the docker file system. In practice it means it should be a subdirectory of the current directory.\n\n### Tests\n\nA test cogment app is defined in `./tests/test_cogment_app`. To make things easier, generated files are versioned in the repository. To get a fresh/updated generation, simply run\n\n```\nCOGMENT_PATH=/path/to/your/cogment poetry run task generate_test_cogment_app\n```\n\n#### Module tests\n\nThese tests only rely on the sdk, no connection to an orchestrator is done.\n\nTo execute the module tests, simply run\n\n```\npoetry run task test\n```\n\n#### Integration tests\n\nThese tests launch and use an orchestrator they are slower but more in depth. To run them the first step is to configure the way to launch the orchestrator in a `.env` file. You can copy `.env.template` for an example of what\'s expected.\n\nThen, to execute the integration tests (as well as the module tests), simply run\n\n```\npoetry run task test --launch-orchestrator\n```\n\nThese tests can also be launched in a docker image.\n\n```\ndocker build -t cogment/cogment-py-sdk-integration-test:latest --build-arg COGMENT_ORCHESTRATOR_IMAGE="<PATH_TO_COGMENT_ORCHESTRATOR_IMAGE" -f integration_test.dockerfile .\ndocker run --rm cogment/cogment-py-sdk-integration-test:latest\n```\n\n### Lint\n\nRun the linter using\n\n```\npoetry run task lint\n```\n\n### Build the source package\n\nBuild the source package (this step will only be succesfull if `poetry install` succeeded)\n\n```\npoetry build -f sdist\n```\n\n### Build a Docker image\n\nNavigate to the python SDK directory and run the following in order to create an image that can be used by a cogment project:\n\n```\ndocker build -t image_name .\n```\n\n\n',
    'author': 'Artificial Intelligence Redefined',
    'author_email': 'dev+cogment@ai-r.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://cogment.ai',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}
from build import *
build(setup_kwargs)

setup(**setup_kwargs)



# Copyright 2020 Artificial Intelligence Redefined <dev+cogment@ai-r.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from cogment.api.data_pb2_grpc import LogExporterServicer
from cogment.api.data_pb2 import LogExporterSampleReply
from cogment.trial import Trial
from cogment.datalog import _ServedDatalogSession
from cogment.utils import raw_params_to_user_params, list_versions
import logging
import asyncio


class LogExporterService(LogExporterServicer):

    def __init__(self, impl, cog_settings):
        self.__impl = impl
        self.__cog_settings = cog_settings
        logging.info("Log Exporter Service started")

    async def OnLogSample(self, request_iterator, context):
        metadata = dict(context.invocation_metadata())
        trial_id = metadata["trial-id"]

        trial = Trial(trial_id, [], self.__cog_settings)

        msg = await request_iterator.__anext__()
        assert msg.HasField("trial_params")
        trial_params = raw_params_to_user_params(msg.trial_params, self.__cog_settings)

        session = _ServedDatalogSession(self.__impl, trial, trial_params)
        loop = asyncio.get_running_loop()
        session._task = loop.create_task(session._run())

        # TODO: Wait for session._started
        async for msg in request_iterator:
            if session._task.done():
                break
            assert msg.HasField("sample")
            session._new_sample(msg.sample)

        if not session._task.done():
            session._end()

        return LogExporterSampleReply()

    async def Version(self, request, context):
        try:
            return list_versions()
        except Exception:
            traceback.print_exc()
            raise
